
# 自作関数・WebSocket・OpenAI連携システム 解説書

このドキュメントは、WebSocket を利用した音声認識システムと OpenAI API との連携方法について解説します。  
クライアント（HTML/JavaScript）側とサーバー（FastAPI/Python）側の設定、動作確認、細かい UI 制御の実装方法など、各ステップを順を追って説明しています。

## 目次

1. [前提条件](#前提条件)
2. [WebSocketの疎通確認](#websocketの疎通確認)
   - [HTML側の設定](#html側の設定)
   - [Python側の設定](#python側の設定)
   - [環境設定とテスト](#環境設定とテスト)
3. [Step2. 音声認識機能の追加](#step2-音声認識機能の追加)
   - [HTML側の設定](#html側の設定-音声認識)
   - [テスト手順](#テスト手順)
4. [OpenAIとのソケット接続](#openaiとのソケット接続)
   - [HTML側の修正](#html側の修正)
   - [Python側の修正とエンドポイント追加](#python側の修正とエンドポイント追加)
   - [テスト手順](#テスト手順-2)
5. [細かい制御の追加](#細かい制御の追加)
   - [HTML側での操作制御とオーバーレイ表示](#html側での操作制御とオーバーレイ表示)
   - [サーバーからの応答受信処理](#サーバーからの応答受信処理)
6. [まとめ](#まとめ)

---

## 前提条件

- **実行環境**
  - Python 3.x
  - FastAPI (および関連パッケージ：uvicorn、python-dotenv など)
- **APIキー**
  - OpenAI API キーをプロジェクトルートの `.env` ファイルに設定すること
- **必要ファイル**
  - `myopenai_realtime.py` および `myopenai.py` （サーバー側の処理用モジュール）
  - クライアント側 JavaScript ファイル（`AudioClient.js` または `AudioController.js` ※ファイル名の整合性に注意）


---

## WebSocketの疎通確認

### HTML側の設定

1. **ファイル配置**  
   - JavaScript ファイル（例: `AudioClient.js` または `AudioController.js`）を `static` フォルダに配置してください。  
     ※以降、統一したファイル名（例：`AudioClient.js`）を使用する場合は、すべて同じ名前にしてください。

2. **index.html の作成**  
   以下の内容で `static/index.html` を作成します。

   ```html
   <!DOCTYPE html>
   <html lang="ja">
   <head>
       <meta charset="UTF-8">
       <title>WebSocket接続確認</title>
       <!-- ※読み込むJSファイル名が実際のファイルと一致しているか確認 -->
       <script src="static/AudioClient.js"></script>
   </head>
   <body>
       <h1>WebSocket接続テストページ</h1>
   </body>
   </html>
   ```

### Python側の設定

1. **ファイル配置と FastAPI の設定**  
   以下は `main.py` のサンプルコードです。  
   ※`myopenai_realtime.py` と `myopenai.py` の実装が既に用意されている前提です。

   ```python
   import os, uuid
   import asyncio
   from fastapi import FastAPI, Query, WebSocket, Body
   from fastapi.staticfiles import StaticFiles
   from fastapi.responses import HTMLResponse
   from myopenai_realtime import myopenai_realtime
   from dotenv import load_dotenv

   load_dotenv()

   app = FastAPI()
   app.mount("/static", StaticFiles(directory="static"), name="static")

   # クライアントごとのハンドラー管理用辞書
   mo_handlers = {}
   API_KEY = os.getenv("OPENAI_API_KEY")

   # --------------------------------------------------------------
   # トップページの提供
   # --------------------------------------------------------------
   @app.get("/", response_class=HTMLResponse)
   async def read_index():
       with open("static/index.html", "r", encoding="utf-8") as f:
           return f.read()

   # --------------------------------------------------------------
   # WebSocket接続エンドポイント
   # --------------------------------------------------------------
   @app.websocket("/ws")
   async def websocket_endpoint(websocket: WebSocket):
       await websocket.accept()
       print(f"クライアントとWebSocket接続しました: {websocket.client}")

       # オーディオ処理用ハンドラーの生成（初期設定：whisper モード有効）
       audio_handler = myopenai_realtime(API_KEY, websocket, f_whispermode=True)
       client_id = audio_handler.get_client_id()
       mo_handlers[client_id] = audio_handler
       await audio_handler.send_client_id()

       # クライアントからのデータ受信（音声など）を待つ
       await audio_handler.handle_client_websocket()

       print(f"クライアント切断: {client_id}")
       # 切断時に OpenAI との接続もクローズ
       await mo_handlers[client_id].close_openai_ws()
       del mo_handlers[client_id]
   ```

### 環境設定とテスト

1. **環境設定**  
   プロジェクトルートに `.env` ファイルを作成し、以下のように OpenAI API キーを設定します：

   ```
   OPENAI_API_KEY=your_openai_api_key_here
   ```

2. **テスト手順**  
   - サーバー（`main.py`）を起動する（例: `uvicorn main:app --reload`）。
   - ブラウザで `http://localhost:8000/` にアクセス。
   - サーバーコンソールに「クライアントとWebSocket接続しました」と表示され、  
     クライアント側のコンソールに `client_id` が出力されれば接続成功です。

---

## Step2. 音声認識機能の追加

### HTML側の設定（音声認識用UI）

1. **UI要素の配置**  
   以下の要素を配置することで、録音開始・停止ボタンと発話内容・AI応答の表示領域を実現します。

2. **サンプルコード**

   ```html
   <!DOCTYPE html>
   <html lang="ja">
   <head>
       <meta charset="UTF-8">
       <meta name="viewport" content="width=device-width, initial-scale=1.0">
       <title>録音アプリ</title>
       <!-- AudioClient.js を読み込む -->
       <script src="static/AudioClient.js"></script>
   </head>
   <body>
       <h1>録音アプリ</h1>
       <button id="startButton">録音開始</button>
       <button id="stopButton" style="display: none;">録音停止</button>

       <h2>自分の話した内容</h2>
       <div id="userSpeech"></div>

       <h2>AIの回答</h2>
       <div id="aiResponse"></div>
   </body>
   <script>
       document.addEventListener("DOMContentLoaded", function() {
           const startButton = document.getElementById('startButton');
           const stopButton = document.getElementById('stopButton');

           // 録音開始ボタン押下時の処理
           startButton.addEventListener('click', () => {
               startButton.style.display = 'none';  // 録音開始ボタンを非表示
               stopButton.style.display = 'block';  // 録音停止ボタンを表示
               window.audioClient.startAudioStreaming();
           });

           // 録音停止ボタン押下時の処理
           stopButton.addEventListener('click', () => {
               stopButton.style.display = 'none';  // 録音停止ボタンを非表示
               startButton.style.display = 'block';  // 録音開始ボタンを表示
               window.audioClient.stopAudioStreaming();
           });

           // AudioClient の設定
           const config = {
               WS_URL: "/ws",
               INITIAL_SIGNAL_URL: "/initial-signal",
               DISCONNECT_SIGNAL_URL: "/disconnect-signal",
               START_SIGNAL_URL: "/start-signal",
               STOP_SIGNAL_URL: "/stop-signal",
               AUDIO_SEND_SIGNAL: "input_audio_buffer.append",
               STOP_WAIT_MS: 500,
           };
           window.audioClient = new AudioClient(config);
           window.audioClient.connect_websocket();
       });

       // ユーザー発話・AI回答表示用の要素
       const userSpeech = document.getElementById('userSpeech');
       const aiResponse = document.getElementById('aiResponse');
   </script>
   </html>
   ```

### テスト手順

- ページを読み込んだ後、**録音開始**ボタンを押してマイクに向かって話してみてください。  
- Whisper による音声認識が行われ、ブラウザのコンソールおよび Python コンソールに認識結果が表示されれば成功です。

---

## OpenAIとのソケット接続

ユーザーの発話に基づいて OpenAI から AI 応答を取得するため、クライアントとサーバー間の追加シグナル送受信機能を実装します。

### HTML側の修正

- `document.addEventListener` 内の最後に以下のコードを追加し、OpenAI との接続初期化シグナルを送信します：

  ```javascript
  window.audioClient.sendSignal_initialize();
  ```

### Python側の修正とエンドポイント追加

1. **AudioHandler の初期化変更**  
   OpenAI 連携時には、チャットモードを有効（`f_chatmode=True`）、Whisper モードを無効（`f_whispermode=False`）に変更します。  
   例:

   ```python
   audio_handler = myopenai_realtime(API_KEY, websocket, f_chatmode=True, f_whispermode=False)
   ```

2. **新規エンドポイントの追加**

   OpenAI との接続開始と切断のため、以下のエンドポイントを `main.py` に追加します。

   ```python
   # OpenAIとのソケットを開くエンドポイント
   @app.post("/initial-signal/{client_id}")
   async def initialize_signal(client_id: str):
       # 例: 「日本語で答えてください。」という指示を送信
       return await mo_handlers[client_id].connect_and_session_update("日本語で答えてください。", "echo")

   # OpenAIとのソケットを閉じるエンドポイント
   @app.post("/disconnect-signal/{client_id}")
   async def disconnect_signal(client_id: str):
       return await mo_handlers[client_id].close_openai_ws()
   ```

3. **/ws エンドポイントの修正**  
   クライアント切断時に、OpenAI とのソケット接続も必ずクローズするよう、以下のコードを追加してください（`del mo_handlers[client_id]` の前など）。

   ```python
   print(f"クライアント切断: {client_id}")
   await mo_handlers[client_id].close_openai_ws()
   ```

### テスト手順

- ブラウザから「こんにちは。聞こえてますか？」などと話してみてください。  
- Python コンソールに OpenAI からの返答メッセージが表示され、クライアント側でも応答音声が再生されれば成功です。

---

## 細かい制御の追加

ユーザー操作を適切に制御し、接続待ち状態や AI 応答中の操作を防ぐための工夫を実装します。

### HTML側での操作制御とオーバーレイ表示

1. **接続待機中のボタン無効化**  
   `sendSignal_initialize()` 呼び出し後、OpenAI との接続が確立するまで録音開始ボタンを無効にします。

   ```javascript
   startButton.disabled = true;
   const checkConnection = setInterval(() => {
       if (window.audioClient.is_connect_openai) {
           clearInterval(checkConnection);
           startButton.disabled = false; // 接続が確立されたらボタンを有効化
       }
   }, 100);
   ```

2. **オーバーレイ表示による操作制限**  
   AI の回答中または回答再生中は、ユーザーが他の操作を行えないようオーバーレイを表示します。  
   `<body>` 内に以下の要素を追加してください：

   ```html
   <!-- オーバーレイ要素 -->
   <div id="overlay" style="display: none; position: fixed; top: 0; left: 0; width: 100%; height: 100%; background-color: rgba(0, 0, 0, 0.5); z-index: 1000;"></div>
   ```

   そして、以下のスクリプトで状態を監視します：

   ```javascript
   // isPlaying または isResponding の状態によりオーバーレイ表示を制御
   const overlay = document.getElementById('overlay');
   function monitorAudioPlayback() {
       if (window.audioClient.isPlaying || window.audioClient.isResponding) {
           overlay.style.display = 'block';
       } else {
           overlay.style.display = 'none';
       }
   }
   setInterval(monitorAudioPlayback, 100);
   ```

### サーバーからの応答受信処理

サーバー（または OpenAI）からの応答を受信するため、`responseQueue` を監視し、受信したメッセージを適宜処理するコード例です：

```javascript
// 100msごとに responseQueue をチェック
setInterval(() => {
    if (window.audioClient.responseQueue.length > 0) {
        const response = window.audioClient.responseQueue.shift(); // キューからメッセージを取り出す
        console.log("新しいレスポンス:", response);
        // 必要に応じて、UI の更新処理などを実装
    }
}, 100);
```

---

## まとめ

以上の手順により、以下の機能を段階的に実装できます：

- **WebSocket 通信の確立**  
  クライアントとサーバー間での接続確認、各クライアントごとのハンドラー管理

- **音声認識機能**  
  録音開始・停止ボタン、ユーザーの発話認識、画面表示の実装

- **OpenAI 連携**  
  初期シグナル送信による OpenAI API 接続、チャットモードでの応答取得

- **UI 制御**  
  接続待機中のボタン無効化、AI 応答中のオーバーレイ表示、応答キューの監視

各機能を実装後、段階的にテストを行いながら動作確認してください。  
また、ファイル名やパスの不整合、環境設定のミスに注意し、プロジェクトに合わせてコードを適宜修正してください。

---

この解説書がプロジェクトの参考になれば幸いです。  
必要に応じてさらに機能を拡張する場合は、適宜このドキュメントを更新してください。
